"""Message serialization (messagepack by default).
"""

import msgpack


class Packer(object):
    def pack(self, data):
        return msgpack.packb(data, encoding='utf-8', use_bin_type=True)

    def unpack(self, data):
        return msgpack.unpackb(data, encoding='utf-8', use_list=False)



"""zmq-rpc implementation with pyzmq.
This file contains stuff common to both the client and the server.
"""

from packer import Packer


class ZmqRpcConnector(object):
    """Base class for zmq-rpc client & server implementations with pyzmq.
    """

    def __init__(self):
        self.packer = Packer()

    def send(self, data):
        """Send data to the connected peer.
        :type data: List.
        """

        # Send the data as multipart (that is what aiozmq uses).
        return self.socket.send_multipart(data)

    def receive(self):
        """Receive data sent by the connected peer.
        :rtype: List.
        """

        # Receive the data as multipart (that is what aiozmq uses).
        return self.socket.recv_multipart()



# -*- encoding: utf-8 -*-

import logging
import os
import random
import struct
import time
import zmq

from base import ZmqRpcConnector
from errors import ZmqRpcError


logger = logging.getLogger(__name__)
log = logger.debug


class ZmqRpcClient(ZmqRpcConnector):
    """zmq-rpc client implementation with pyzmq.
    """

    # Taken from aiozmq/rpc/rpc.py (_ClientProtocol).
    REQ_PREFIX = struct.Struct('=HH')
    REQ_SUFFIX = struct.Struct('=Ld')
    RESP = struct.Struct('=HHLd?')

    def __init__(self, url, timeout=None):
        """Initialize the connector.
        :param url: URL to the aiozmq server.
        :param timeout: Timeout when waiting for the server before throwing an
        exception (in milliseconds - use "None" for no timeout).
        """

        super(ZmqRpcClient, self).__init__()

        # Taken from aiozmq/rpc/rpc.py (_ClientProtocol.__init__).
        self.counter = 0
        self.prefix = self.REQ_PREFIX.pack(
            os.getpid() % 0x10000, random.randrange(0x10000)
        )

        # Open a regular 0mq connection, in "dealer" mode.

        self.context = zmq.Context().instance()
        if timeout is not None:
            self.context.RCVTIMEO = timeout
        self.socket = self.context.socket(zmq.DEALER)
        self.socket.connect(url)

    def __getattr__(self, name):
        """Allow directly calling RPC functions.
        """

        return lambda *args, **kwargs: self.callFunction(name, args, kwargs)

    def callFunction(self, func_name, func_args, func_kwargs):
        """Call the specified RPC-exposed function and return its result
        (blocking).

        :param func_name: The name of the RPC function exposed by the server.
        :type func_name: String.

        :param func_args: The positional arguments to send the RPC function.
        :type func_args: List.

        :param func_kwargs: The named arguments to send the RPC function.
        :type func_kwargs: Dictionary.

        :raise zmq_rpc.errors.ZmqRpcError.

        :return: Result of the RPC function call.
        """

        # Example data: [
        #     '\x00k\x8bEg',
        #     '\xf1BO\xd4\x01\x00\x00\x00\r\x15\xd3r\xc2?\xd5A',
        #     'test_func',
        #     '\x92\x01\x02',
        #     '\x80'
        # ]

        data = [
            self._newHeader(),
            func_name.encode('utf-8'),
            self.packer.pack(func_args),
            self.packer.pack(func_kwargs),
        ]

        self.send(data)
        log('sent data: %s', data)

        data = self.receive()
        log('received data: %s', data)

        # Interpret the received data.
        # Taken from aiozmq/rpc/rpc.py (_ClientProtocol.msg_received).
        header, answer = data
        pid, rnd, req_id, timestamp, is_error = self.RESP.unpack(header)
        answer = self.packer.unpack(answer)

        # Handle errors.
        if is_error:
            raise ZmqRpcError(answer[2])  # Just the repr.

        return answer

    def close(self):
        """Close the connection.
        """

        self.socket.close()

    def _newHeader(self):
        # Taken from aiozmq/rpc/rpc.py (_ClientProtocol._new_id).
        self.counter += 1
        if self.counter > 0xffffffff:
            self.counter = 0
        return self.prefix + self.REQ_SUFFIX.pack(self.counter, time.time())


def main():
    """Start a test client.
    """

    logging.basicConfig(
        format='%(asctime)s - %(levelname)s - %(message)s',
        level=logging.DEBUG,
    )

    log('start')

    # The client doesn't have to run in another thread but this is just a test
    # to ensure multi-threaded stuff works as expected.

    def runClient():
        zmq_rpc_client = ZmqRpcClient('tcp://127.0.0.1:2678', timeout=1000)

        call_result = zmq_rpc_client.test_func(
            1, 2, zmq_rpc_client.packer.pack('abc')
        )
        log('result of the test_func call: %s', call_result)

        zmq_rpc_client.close()
        log('closed')

    import threading
    thread = threading.Thread(target=runClient)
    thread.start()
    thread.join()

    log('done')


if __name__ == '__main__':
    main()



# flake8: noqa

from zmq_rpc.client import ZmqRpcClient

from zmq_rpc.errors import ZmqRpcError

from zmq_rpc.packer import Packer

from zmq_rpc.server import RpcMethod
from zmq_rpc.server import ZmqRpcServer



import codecs
from setuptools import find_packages
from setuptools import setup


with codecs.open('README.rst', 'r', 'utf-8') as f:
    README = f.read()

with codecs.open('CHANGES.rst', 'r', 'utf-8') as f:
    CHANGES = f.read()

with codecs.open('CONTRIBUTORS.rst', 'r', 'utf-8') as f:
    CONTRIBUTORS = f.read()


requires = [
    'msgpack-python',
    'pyzmq',
]


setup(
    name='zmq_rpc',
    version='0.2',
    description='0mq RPC calls implemented with pyzmq.',
    long_description=u"{}\n{}\n{}".format(README, CONTRIBUTORS, CHANGES),
    classifiers=[
        'Programming Language :: Python',
    ],
    author='XCG Consulting',
    author_email='contact@xcg-consulting.fr',
    url='http://odoo.consulting',
    keywords='zmq rpc msgpack',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=requires,
    tests_require=requires,
    test_suite='zmq_rpc',
)



# -*- encoding: utf-8 -*-

import zmq
import struct
import logging
import os
import random
import sys
import time
from base import ZmqRpcConnector


logger = logging.getLogger(__name__)
log = logger.debug


class RpcMethod(object):
    """Decorator to mark a function as a RPC method, and optionally give it
    a name that can include periods and Unicode characters.
    Use: @RpcMethod or @RpcMethod(name)
    """

    def __new__(cls, name_or_func=None):
        if name_or_func is not None and hasattr(name_or_func, '__call__'):
            new = super(RpcMethod, cls).__new__(cls, None)
            new.__init__()
            return new.__call__(name_or_func)
        else:
            new = super(RpcMethod, cls).__new__(cls, name_or_func)
            return new

    def __init__(self, name=None):
        self.name = name

    def __call__(self, func):
        name = self.name if self.name is not None else func.__name__
        func.__rpc__ = name
        return func


class ZmqRpcServerMeta(type):
    """Add all methods that have been marked as RPC in the class definition
    (with the :class:`RpcMethod` decorator) to the class' RPC interface.
    """

    def __init__(cls, name, bases, nmspc):

        if '_methods' not in nmspc:
            cls._methods = nmspc['_methods'] = {}
        for attr, value in nmspc.iteritems():
            rpc_name = getattr(value, '__rpc__', None)
            if rpc_name:
                cls._methods[rpc_name] = getattr(cls, attr)

        super(ZmqRpcServerMeta, cls).__init__(name, bases, nmspc)


class ZmqRpcServer(ZmqRpcConnector):
    """zmq-rpc server implementation with pyzmq.
    """

    __metaclass__ = ZmqRpcServerMeta
    _methods = {}

    REQ = struct.Struct('=HHLd')
    RESP_PREFIX = struct.Struct('=HH')
    RESP_SUFFIX = struct.Struct('=Ld?')

    def __init__(self, url):
        """Initialize the connector.
        :param url: URL for the aiozmq server.
        """

        super(ZmqRpcServer, self).__init__()

        self.prefix = self.RESP_PREFIX.pack(os.getpid() % 0x10000,
                                            random.randrange(0x10000))
        self.running = False

        # Open a regular 0mq connection, in "dealer" mode.
        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.ROUTER)
        self.socket.bind(url)

    def rpc_call(self, data):
        """Interpret the received data to call the specified method and
        return its result.

        :param: data: the RPC function call sent by a client.
        :type: data: list

        :return: The formatted result of the RPC function call.
        """
        try:
            pre = data[:-4]
            header, bname, bargs, bkwargs = data[-4:]
            pid, rnd, req_id, timestamp = self.REQ.unpack(header)

            name = bname.decode('utf-8')
            args = self.packer.unpack(bargs)
            kwargs = self.packer.unpack(bkwargs)
        except Exception as exc:
            logger.critical("Cannot unpack %r", data, exc_info=sys.exc_info())
            return
        try:
            func = self._methods[name]
            res = func(self, *args, **kwargs)
            return self.rpc_result(res, req_id, pre, False)
        except Exception as exc:
            exc_info = (
                exc.__class__.__module__ + '.' + exc.__class__.__name__,
                exc.args, repr(exc)
            )
            return self.rpc_result(exc_info, req_id, pre, True)

    def rpc_result(self, ret, req_id, pre, exc):
        """Format the result as a RPC call reply.

        :param: ret: the return value of the RPC call.

        :param: req_id: the identifier of the request.
        :type: req_id: list

        :param: pre: the unchanged prefixes from the request.
        :type: pre: list

        :param: exc: whether the result signals an exception.
        :type: exc: bool

        :return: The formatted result of the RPC function call.
        """
        prefix = self.prefix + self.RESP_SUFFIX.pack(req_id, time.time(), exc)
        return pre + [prefix, self.packer.pack(ret)]

    def run(self):
        """Start the server execution loop.
        """

        self.running = True
        while self.running:
            data = self.receive()
            reply = self.rpc_call(data)
            self.send(reply)

    def stop(self):
        """Stop the execution loop after the current RPC call has been treated.
        """
        self.running = False


def main():
    """Start a test server.
    """

    logging.basicConfig(
        format='%(asctime)s - %(levelname)s - %(message)s',
        level=logging.DEBUG,
    )

    class AddServer(ZmqRpcServer):
        @RpcMethod
        def test_func(self, a, b):
            self.stop()
            return a + b

    log('start')

    zmq_rpc_server = AddServer('tcp://*:2678')
    zmq_rpc_server.run()

    log('done')


if __name__ == '__main__':
    main()
